/*
    hexutil.c - massage Intel hex and Motorola S-record files

Options:   -Sn     set output format to Sn records (default S1)
           -I      set output format to Intel records
           -Rnnnn  relocate addresses by nnn
           -Lnnnn  set lower limit of output addresses to nnnn
           -Hnnnn  set upper limit of output addresses to nnnn - 1
           -E      output data at even addresses only
           -O      output data at odd addresses only
           -Z      always output records with address of zero

           -L and -H are not compatible with -E and -O
           -I is not compatible with -Sn

   Other arguments are interpreted as input file and output file resp.

   Processing order:    relocation, { low/high clipping, even/odd filtering }

   Author:     Harold Rabbie, Echelon, November 1994

*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#define BUFSIZE 256

typedef unsigned long   address;
typedef unsigned char   data;
typedef enum { FALSE, TRUE } boolean;

void    get_options( int argc, char * argv[ ] );
void    start_rec( int );
void    put_data( data );
boolean get_data( data * );
void    end_rec( address, int type );
boolean read_rec( void );
void    slice_record( void );
void    clip_record( void );
boolean process_record( boolean intel );
void    bad_record( const char * msg );
int     addr_length( int );
data    addrs2xsum( address );
void    error( const char * str, const char * arg );
void    usage( void );
char  * switch_val( char * );

data    inbuffer[ BUFSIZE / 2 ];    // input unpacked data
data  * inbuf_ptr;          // next byte to read

//      Command line options and defaults

address relocation      = 0;
address low_lim     = 0;
address high_lim    = 0;    // 0 means no high limit
boolean address_zero    = FALSE;
boolean even_only       = FALSE;
boolean odd_only    = FALSE;
int     rec_type    = 1;    // 0 means intel
const int RESET_type  = -1;   // Neuron Chip reset record
#define MOTOROLA (rec_type != 0)

address out_address;    // address of current S-record
int     out_data_count; // number of data bytes in output S-record
int     in_data_count;  // number of data bytes in input S-record
data    inbuffer[ BUFSIZE / 2 ];    // input unpacked data
data    *inbuf_ptr;                 // next byte to read


//      MAIN PROGRAM LOOP

void main( int argc, char * argv[] )
{
    char    *temp;
    int     type;

// begin
    get_options( argc, argv );

    if( MOTOROLA ) {     // motorola
        start_rec( 0 );      // initial S0 record to output
        for( temp = "(C) Echelon"; *temp; temp++ ) put_data( *temp );
        end_rec( 0, 0 );
    }

    while( read_rec() ) {      // until end-of-file
        out_address += relocation;
        inbuf_ptr = inbuffer;
        if( even_only || odd_only) slice_record();
        else                       clip_record();
    }

    type = MOTOROLA ? ( 10 - rec_type ) : 1;     // type for EOF record
    start_rec( type );  // final record to output
    end_rec( 0, type );
    exit( 0 );
}       // end main

void clip_record( void )   // output data clipped against low/high limits
{
    address first_addr = 0;
    data    data_value;
    int     type;

    // begin
    // FIXUP - need quick clip here, but WATCH FOR WRAPAROUND!

    start_rec( rec_type );      // start data rec
    while( get_data( &data_value ) ) {
        if( out_address >= low_lim )
        if( !high_lim || ( out_address < high_lim ) ) {

            if( !out_data_count ) first_addr = out_address;
            put_data( data_value );
        }
        out_address++;
    }
    if( ! out_address && address_zero ) type = RESET_type;
    else type = rec_type;

    end_rec( first_addr, type );  // end data rec
}       // end clip_record

void slice_record( void )  // output data sliced even or odd
{
    data    data_value;
    boolean output_on;      // toggle for outputting data

// begin
    start_rec( rec_type );      // start data rec
    output_on = ( out_address & 1 ) ? odd_only : even_only;
    out_address /= 2;
    if( !output_on && even_only ) out_address++;   // don't ask, it works

    while( get_data( &data_value ) ) {

        if( output_on ) put_data( data_value );
        output_on = !output_on;
    }
    end_rec( out_address, rec_type ); // end data rec
}       // end slice_record

data addrs2xsum( address addr ) // compute 8-bit checksum of 32-bit value
{
    data   * xsum;
    xsum = ( data * ) & addr;
    return * xsum++ + * xsum++ + * xsum++ + * xsum;
}

//      INPUT SECTION

char    in_record[ BUFSIZE ];          // input S-record
char    *in_record_ptr;            // next char to process

boolean read_rec( void )      // read S-record from input
                // return FALSE on EOF
{
    char * status;

    // begin

    for(;;) {
        status = fgets( in_record, BUFSIZE, stdin );
        in_record_ptr = in_record;
        if( !status ) return( FALSE );
        switch( * in_record_ptr++ ) {
            case '\n' : continue;
            case 'S'  : if( ! process_record( FALSE ) ) continue;
                    return( TRUE );
            case ':'  : if( ! process_record( TRUE ) ) continue;
                    return( TRUE );
            default   : bad_record( "Illegal first char" );
                    continue;
        }
    }
}       // end read_rec

boolean process_record( boolean intel )
                        // read and process the rest of the record
                        // return TRUE if record was processed
{
    int     type;       // record_type
    int     count;      // count field from S record
    data    checksum;
    char    format[ 10 ];
    int     i;
    int     length;     // length of address in characters
    int     temp;

// begin
    if( intel ) {
        if( sscanf( in_record_ptr, "%2x%4lx%2x", &count, &out_address,
        &type ) != 3 ) {
        bad_record( "Invalid count, address, type" );
        return FALSE;
        }
        if( type ) return FALSE;      // ignore type != 0
        length = 4;           // intel address length
        in_data_count = count;    // intel data count
        in_record_ptr += 8;

    } else {    // motorola
        if( sscanf( in_record_ptr++, "%1d", & type ) != 1 ) {
        bad_record( "Invalid type" );
        return FALSE;
        }

        if( type < 1 || type > 3 ) return FALSE;    // ignore S0, S4-S9

        length = addr_length( type );

        sprintf( format, "%%2x%%%1dlx", length );
        if( sscanf( in_record_ptr, format, &count, &out_address ) != 2 ) {
        bad_record( "Invalid count, address" );
        return FALSE;
        }
        in_record_ptr += length + 2;
        in_data_count = count - length / 2 - 1;
    }       // end if( intel )

    if( in_data_count >= BUFSIZE / 2 ) {
        bad_record( "Record too long" );
        return FALSE;
    }

    checksum = addrs2xsum( out_address ) + count;

    for( i = 0, inbuf_ptr = inbuffer; i <= in_data_count; i++ ) {
        // read xsum field, too
        if( sscanf( in_record_ptr, "%2x", &temp ) != 1 ) {
            bad_record( "Invalid hex data" );
            return FALSE;
        }
        *inbuf_ptr++ = temp;
        checksum += temp;
        in_record_ptr += 2;
    }

    if( checksum != ( intel ? 0 : 0xff ) ) {
        bad_record( "Checksum error" );
        return FALSE;
    }
    return TRUE;
}       // process_record

void bad_record( const char * msg )
{
    fprintf( stderr, "Bad record: %s %s\n", msg, in_record );
    in_data_count = 0;
}       // end bad_record

boolean get_data( data * data_ptr )    // get next byte from input record
                // return FALSE when no more data
{
    if( ! in_data_count-- ) return( FALSE );
    *data_ptr = *inbuf_ptr++;
    return( TRUE );
}       // end get_data


//      OUTPUT SECTION

char    outbuffer[ BUFSIZE ];
data    checksum;
char    *outbuf_ptr;

int addr_length( int type )     // get length of address in characters
{
    switch( type ) {
        case 0: if( MOTOROLA ) return( 0 ); else return( 4 );
        case 1: case 9: return( 4 );
        case 2: case 8: return( 6 );
        case 3: case 7: return( 8 );
    }
    return 0;
}       // end addr_length

void start_rec( int type )       // begin output hex record
{
    outbuf_ptr = outbuffer;
    out_data_count = 0;
    checksum = 0;

    if( MOTOROLA ) {       // motorola
        *outbuf_ptr++ = 'S';
        *outbuf_ptr++ = '0' + type;
        outbuf_ptr += 2 + addr_length( type );  // skip count and address
    } else {               // intel
        *outbuf_ptr++ = ':';
        outbuf_ptr += 6;    // skip count and address
        sprintf( outbuf_ptr, "%02X", type );
        outbuf_ptr +=2;
    }
}       // end start_rec

void put_data( data data_value )      // put data byte into output S-record
{
    sprintf( outbuf_ptr, "%02X", data_value );
    outbuf_ptr += 2;
    out_data_count++;
    checksum += data_value;
}       // end put_data

void end_rec( address start_addr, int type )       // end output S-record
{
    int     count;
    char    temp;
    int     length;
    static const address length_table[ ]
        = { 0x1UL, 0x100UL, 0x10000UL, 0x1000000UL };
// begin
    if( !out_data_count ) {     // no data
        if( MOTOROLA ) {        // motorola
            if( type >= 1 && type <= 3 ) return;
        } else {        // intel
            if( type == 0 ) return;
        }
    }
    if( type == RESET_type ) type = rec_type;     // special case

    if( MOTOROLA ) count = ( outbuf_ptr - outbuffer - 2 ) / 2;  // moto
    else count = ( outbuf_ptr - outbuffer - 9 ) / 2;    // intel

    checksum += addrs2xsum( start_addr ) + count;
    if( !MOTOROLA ) checksum += type;

    sprintf( outbuf_ptr, "%02X",
        ( ( MOTOROLA ? -1 : 0 ) - checksum ) & 0xff );

    length = addr_length( type );
    if( length < 8 && start_addr >= length_table[ length * 2 ] ) {
         fprintf( stderr, "Address out of range: %X\n", start_addr );
         return;
    }

    if( MOTOROLA ) {            // motorola
        temp = outbuffer[ length + 4 ];
            // this char will be zapped by sprintf
        if( type )
        sprintf( &outbuffer[ 2 ], "%02X%0*X", count, length, start_addr );
        else sprintf( &outbuffer[ 2 ], "%02X", count );

        outbuffer[ length + 4 ] = temp;     // restore char
    } else {            // intel
        temp = outbuffer[ 7 ];
            // this char will be zapped by sprintf
        sprintf( &outbuffer[ 1 ], "%02X%04X", count, start_addr );

        outbuffer[ 7 ] = temp;     // restore char
    }
    puts( outbuffer );

}       // end end_rec

//      COMMAND LINE PROCESSING SECTION

char * switch_val( char * argv_ptr )
{
    if( ! * argv_ptr ) usage();
    return ++ argv_ptr;
}

void get_options( int argc, char * argv[] )
{
    char  * argv_ptr;
    int     arg;
    boolean motorola = FALSE;
    boolean intel    = FALSE;

// begin
    for( arg = 1; arg < argc; arg++ ) {
        argv_ptr = argv[ arg ];
        if( * argv_ptr != '/' && * argv_ptr != '-' ) break;

        switch( toupper( * ++ argv_ptr ) ) {
        case 'R':
            relocation = strtol( switch_val( argv_ptr ), 0, 16 );
            break;
        case 'L':
            low_lim = strtol( switch_val( argv_ptr ), 0, 16 );
            break;
        case 'H':
            high_lim = strtol( switch_val( argv_ptr ), 0, 16 );
            break;
        case 'E':
            even_only = TRUE;
            break;
        case 'O':
            odd_only = TRUE;
            break;
        case 'S':
            rec_type = atoi( switch_val( argv_ptr ) );
            motorola = TRUE;
            break;
        case 'I':
            intel = TRUE;
            rec_type = 0;
            break;
        case 'Z':
            address_zero = TRUE;
            break;
        default:
            usage();
        }   // end switch
    }       // end for

    if( even_only && odd_only )
    error( "Can't specify both -e and -o\n", 0 );

    if( motorola && intel )
    error( "Can't specify both Intel and Motorola", 0 );

    if( intel ) rec_type = 0;
    if( motorola )
    if( rec_type < 1 || rec_type > 3 )
        error( "S-record type must be 1, 2 or 3\n", 0 );

    if( high_lim && ( low_lim > high_lim ) )
    error( "Low limit > high limit\n", 0 );

    if( ( low_lim || high_lim ) && ( even_only || odd_only ) )
    error( "Low/high clipping is not compatible with even/odd slicing\n", 0 );

    if( arg++ >= argc ) return;

    if( ! freopen( argv_ptr, "r", stdin ) )
        error( "Can't open input file %s\n", argv_ptr );

    argv_ptr = argv[ arg ];
    if( arg++ >= argc ) return;

    if( ! freopen( argv_ptr, "w", stdout ) )
        error( "Can't open output file %s\n", argv_ptr );

    if( arg < argc ) error( "Too many arguments\n", 0 );

}       // end get_options

void error( const char * string, const char * arg )
{
    fprintf( stderr, string, arg );
    exit( 1 );
}       // end error

void usage( void )
{
    fprintf( stderr, "Usage: %s %s %s\n %s %s %s\n %s\n",
        "hexutil", "[-Sn] [-I] (Motorola or Intel)", "[-Rrelocation]",
        "[-e], [-o] (even or odd addrs only)",
        "[-Llow_limit]", "[-Hhigh_limit]", "[-Z] (include address zero)" );
    exit( 1 );
}       // end usage
